def get_keras_loss(loss_name):
    """
    Fetches built in keras losses

    :param loss_name: The name of the loss to fetch
    :return: A vanilla Keras loss function
    """

    from keras import losses

    return getattr(losses, loss_name)


def get_custom_loss(loss_function_getter_name, loss_function_getter_params):
    """
    Fetches custom Keras losses

    :param string loss_function_getter_name: The name of the function to used in order to fetch the loss function
    :param dict loss_function_getter_params: The paramters to pass to the getter

    :return: A vanilla Keras loss function
    """

    from mercury_ml.keras.providers import custom_loss_functions

    return getattr(custom_loss_functions, loss_function_getter_name)(**loss_function_getter_params)


import pandas as pd
import numpy as np

class KerasIteratorDataWrapper():
    """
    A base class for DataWrappers that have Keras iterators as their underlying
    """

    def __init__(self, underlying, field_names):
        self.underlying = underlying
        self.field_names = field_names


class KerasIteratorIndexDataWrapper(KerasIteratorDataWrapper):
    """
    A DataWrapper with a Keras DirectoryIterator as its underlying. This particular wrapper is used to access index information (in this case the filenames)
    """

    def to_numpy(self):
        """
        Creates a new NumpyDataWrapper based on this KerasIteratorIndexDataWrapper and returns it as a new object

        :return: a new NumpyDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.numpy import NumpyDataWrapper
        index, index_fields = self._get_numpy_index()
        return NumpyDataWrapper(index, index_fields)

    def to_pandas(self):
        """
        Creates a new PandasDataWrapper based on this KerasIteratorIndexDataWrapper and returns it as a new object

        :return: a new PandasDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.pandas import PandasDataWrapper
        index, index_fields = self._get_numpy_index()
        return PandasDataWrapper(pd.DataFrame(index, columns=index_fields), index_fields)

    def to_keras_directory_iterator(self):
        return self

    def _get_numpy_index(self):
        """
        Converts self.underlying to a numpy array

        :return: index (the new numpy array), index_fields (a list of of headings for the index, in this case just ["filenames"])
        """

        index = np.array(self.underlying.get_filenames())
        index_fields = ["filenames"]

        return index, index_fields

class KerasIteratorTargetsDataWrapper(KerasIteratorDataWrapper):
    """
    A DataWrapper with a Keras DirectoryIterator as its underlying. This particular wrapper is used to access targets information
    """

    def to_numpy(self):
        """
        Creates a new NumpyDataWrapper based on this KerasIteratorTargetsDataWrapper and returns it as a new object

        :return: a new NumpyDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.numpy import NumpyDataWrapper
        targets, targets_fields = self._get_numpy_targets()
        return NumpyDataWrapper(targets, targets_fields)

    def to_pandas(self):
        """
        Creates a new PandasDataWrapper based on this KerasIteratorTargetsDataWrapper and returns it as a new object

        :return: a new PandasDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.pandas import PandasDataWrapper
        targets, targets_fields = self._get_numpy_targets()
        return PandasDataWrapper(pd.DataFrame(targets, columns=targets_fields), targets_fields)

    def to_keras_directory_iterator(self):
        return self

    def _get_numpy_targets(self):
        """
        Converts self.underlying to a numpy array

        :return: targets (the new numpy array), targets_fields (a list of of headings for the targets)
        """

        self.underlying.shuffle = False
        targets = self.underlying.get_target_dummies()
        targets_fields = self.underlying.get_labels_dummies()

        return targets, targets_fields


class KerasIteratorFeaturesDataWrapper(KerasIteratorDataWrapper):
    """
    A DataWrapper with a Keras DirectoryIterator as its underlying. This particular wrapper is used to access features information
    """

    def to_numpy(self):
        """
        Creates a new NumpyDataWrapper based on this KerasIteratorFeaturesDataWrapper and returns it as a new object

        :return: a new NumpyDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.numpy import NumpyDataWrapper
        features, features_fields = self._get_numpy_features()
        return NumpyDataWrapper(features, features_fields)

    def to_pandas(self):
        """
        Creates a new PandasDataWrapper based on this KerasIteratorFeaturesDataWrapper and returns it as a new object

        :return: a new PandasDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.pandas import PandasDataWrapper
        features, features_fields = self._get_numpy_features()
        return PandasDataWrapper(pd.DataFrame(features, columns=features_fields), features_fields)

    def to_keras_directory_iterator(self):
        return self

    def _get_numpy_features(self):
        """
        Converts self.underlying to a numpy array

        :return: features (the new numpy array), None (since field_names are not relevant here)
        """

        self.underlying.batch_size = self.underlying.n
        self.underlying.shuffle = False #TODO check if this is sufficient!

        features, _ = self.underlying.next()

        return features, None


def get_or_create_h2o():
    """
    Gets or initiates and H2O session
    :return:
    """

    import h2o
    h2o.init()
    return h2o

def get_or_create_h2o_sparkling(h2o_context_params=None,  h2o_log_level="ERROR", spark_session_params=None):
    """
    Gets or initiates an H2OSparkling session.

    :param dict h2o_context_params: The parameters based on which the H2OSparkling session is to be initialized
    :param string h2o_log_level: The log level of the H2OSparkling Session
    :param dict spark_session_params: The parameters based on which the Spark session is to be initialized
    :return:
    """

    from pysparkling import H2OConf, H2OContext

    # Start SparkSession
    #TODO possibly change this to create spark session outside and pass "spark" as variable
    from mercury_ml.spark.providers.session import get_or_create_spark_session

    if not spark_session_params:
        spark_session_params = {}

    spark = get_or_create_spark_session(**spark_session_params)

    # Start H2OContext
    h2o_conf = H2OConf(spark)
    h2o_conf.set_h2o_node_log_level(h2o_log_level)

    if not h2o_context_params:
        h2o_context_params = {}

    if h2o_context_params.get("auth"):  # requires h2o-pysparkling>=2.2.28
        h2o_context_params["auth"] = tuple(h2o_context_params["auth"])

    h2o_context = H2OContext.getOrCreate(spark, conf=h2o_conf, **h2o_context_params)

    return h2o_context



def store_hive_table(data, directory, file_name):
    """
    Saves predictions to HIVE

    :param DataFrame prediction_df: The predictions to be saved
    :param string directory: Path where the prediction file should be saved
    :param string file_name: Name that the prediction file should be saved with
    :return: None
    """
    table_name = directory + "." + file_name
    data.write.saveAsTable(table_name)




class DataSet:
    """
    A class that groups together DataWrappers. A typical DataSets would consist of "full_data", "index" , "features" and "targets" DataWrappers.
    """

    def __init__(self, data_wrappers_dict=None):
        if data_wrappers_dict:
            self.add_data_wrappers(data_wrappers_dict)

    def add_data_wrapper(self, data_wrapper_name, data_wrapper):
        """
        Adds a single DataWrapper to this DataSet
        :param string data_wrapper_name: Name of the DataWrapper
        :param DataWrapper data_wrapper: The DataWrapper to add
        :return:
        """
        setattr(self, data_wrapper_name, data_wrapper)

    def add_data_wrappers(self, data_wrappers_dict):
        """
        Adds DataWrappers to this DataSet. For example

        data_set = DataSet()
        data_bunch.add_data_sets(
            {
                "full_data": full_data_data_wrapper,
                "index": index_data_wrapper,
                "features": features_data_wrapper
                "targets": targets_data_wrapper
            }
        )

        :param dict data_sets_dict: Of the form {{"data_wrapper_name": data_wrapper},...}
        :return:
        """

        for data_wrapper_name, data_wrapper in data_wrappers_dict.items():
            self.add_data_wrapper(data_wrapper_name, data_wrapper)

    def add_data_wrapper_via_concatenate(self, left_data_wrapper_name, right_data_wrapper_name, new_data_wrapper_name):
        """
        Concatenates two DataWrappers and add the result as a new DataWrapper

        :param DataWrapper left_data_wrapper_name: The DataWrapper to concatenate to the left.
        :param DataWrapper right_data_wrapper_name: The DataWrapper to concatenate to the right.
        :param string new_data_wrapper_name: The name of the new DataWrapper.
        :return:
        """

        left_data_wrapper = getattr(self, left_data_wrapper_name)
        right_data_wrapper = getattr(self, right_data_wrapper_name)
        self.add_data_wrapper(new_data_wrapper_name, left_data_wrapper.concatenate(right_data_wrapper))

    def transform(self, transformation_params, transform_then_slice=False):
        """
        Transforms of of the underlying DataWrappers to a new type, adds them to a new DataSet and returns that DataSet.
        :param dict transformation_params: The parameters to be used in the DataWrapper transformations
        :param bool transform_then_slice: Determines whether each DataWrapper within the DataSet should be transformed
        independently (False), or whether only "full_data" should be transformed and the subset the derived from there (True)
        :return:
        """

        if transform_then_slice:
            return self._transform_full_data_and_make_slices(**transformation_params)
        else:
            return self._transform(**transformation_params)

    def _transform(self, transform_to, data_wrapper_params=None):
        """
        Transforms each DataWrapper independently
        """

        if data_wrapper_params is None:
            data_wrapper_params =  {k:{} for k in list(self.__dict__.keys())} #TODO: does this work?

        new_data_set = DataSet()
        for data_wrapper_name, params in data_wrapper_params.items():
            current_data_wrapper = getattr(self, data_wrapper_name)
            transform = getattr(current_data_wrapper, "to_" + transform_to)
            new_data_set.add_data_wrapper(data_wrapper_name, transform(**params))

        return new_data_set

    def _transform_full_data_and_make_slices(self, transform_to, full_data_wrapper_params=None, data_wrapper_names=None):
        """
        Transforms only the underlying full data. Populates features, targets and index by slicing the transformed data
        instead of transforming each separately
        """

        current_full_data_wrapper = self.full_data
        transform = getattr(current_full_data_wrapper, "to_" + transform_to)
        new_full_data_wrapper = transform(**full_data_wrapper_params)

        new_data_set = DataSet({"full_data": new_full_data_wrapper})

        if data_wrapper_names is None:
            data_wrapper_names = list(self.__dict__.keys()).remove("full_data")

        for data_wrapper_name, data_wrapper_value in self.__dict__.items():
            if data_wrapper_name in data_wrapper_names:
                field_names = getattr(self, data_wrapper_name).field_names
                new_data_set.add_data_wrappers(
                    {data_wrapper_name: new_full_data_wrapper.__class__(new_full_data_wrapper.slice_on_column_names(field_names), field_names)})

        return new_data_set


def load_hdf5_model(local_dir, filename, extension=".h5", custom_objects=None):
    """
    Loads a Keras model that was saved in ".h5" format

    :param string local_dir: Local directory where the model is saved
    :param string filename: Filename with which the model is saved
    :param string extension: Extension to the filename with which the model is saved
    :param dict custom_objects: Any custom objects  (such as custom loss functions) that were included when the model was saved
    :return: A Keras model
    """
    from keras.models import load_model
    return load_model(local_dir + "/" + filename + extension, custom_objects=custom_objects)



import shutil
import os
import subprocess

def copy_from_disk_to_disk(source_dir, target_dir, filename, overwrite=False, delete_source=False):
    """
    Moves a file from one location of Disk to another location on Disk

    :param string source_dir: Path of file to be copied
    :param string target_dir: Path that file is to be copied to
    :param string filename: Name of file to be copied
    :param bool overwrite: If true, overwrite if file already exists in target_dir
    :param bool delete_source: If true, delete file from source_dir after copying
    :return:
    """

    source_dir = _make_local_path(source_dir)
    target_dir = _make_local_path(target_dir)

    source_path = os.path.join(source_dir, filename)
    target_path = os.path.join(target_dir, filename)

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    if overwrite == False:
        if os.path.exists(target_path):
            pass
            #raise FileExistsError("File already exists, but overwrite was not activated")
        else:
            shutil.copyfile(source_path,target_path)
    else:
        shutil.copyfile(source_path,target_path)

    if delete_source == True:
        # check if copy was successfull:
        if os.path.exists(os.path.join(target_dir, filename)):
            os.remove(os.path.join(source_dir, filename))
        else:
            raise ChildProcessError(
                "Delete Source was activated but copy file to new location failed")


def copy_from_disk_to_hdfs(source_dir, target_dir, filename, overwrite=False, delete_source=False):
    """
    Moves a file from Disk to HDFS

    :param string source_dir: Path of file to be copied
    :param string target_dir: Path that file is to be copied to
    :param string filename: Name of file to be copied
    :param bool overwrite: If true, overwrite if file already exists in target_dir
    :param bool delete_source: If true, delete file from source_dir after copying
    :return:
    """

    # check if directory exists
    exists = subprocess.call(["hadoop", "fs", "-test", "-d", target_dir])

    if exists == 0:  # 0 equals exists
        subprocess.call(["hadoop", "fs", "-copyFromLocal", os.path.join(
            source_dir, filename), os.path.join(target_dir, filename)])
    else:  # create directory first
        subprocess.call(["hadoop", "fs", "-mkdir", "-p", target_dir])
        subprocess.call(["hadoop", "fs", "-copyFromLocal", os.path.join(source_dir, filename),
                         os.path.join(target_dir, filename)])


def copy_from_disk_to_s3(source_dir, target_dir, filename, overwrite=False, delete_source=False, s3_session_params=None,
                         reuse_existing=True):
    """
    Moves a file from Disk to S3

    :param string source_dir: Path of file to be copied
    :param string target_dir: Path that file is to be copied to
    :param string filename: Name of file to be copied
    :param bool overwrite: If true, overwrite if file already exists in target_dir
    :param bool delete_source: If true, delete file from source_dir after copying
    :return:
    """
    if not s3_session_params:
        s3_session_params={}

    import boto3
    if not reuse_existing:
        # s3 = boto3.resource("s3")
        session = boto3.Session(**s3_session_params)
        s3 = session.resource("s3")
    else:
        from mercury_ml.common.providers.artifact_copying import S3Singleton
        s3 = S3Singleton(**s3_session_params).s3

    s3_bucket_name, s3_path = target_dir.split("/", 1)
    s3_key = s3_path + "/" + filename

    source_path = os.path.join(source_dir, filename)
    if overwrite or not _s3_key_exists(s3, s3_bucket_name, s3_key):
        s3.Object(s3_bucket_name, s3_key).put(Body=open(source_path, "rb"))

    if delete_source:
        os.remove(source_path)

def copy_from_disk_to_gcs(source_dir, target_dir, filename, overwrite=False, delete_source=False):
    """
    Moves a file from Disk to GCS (Google Cloud Storage)

    :param string source_dir: Path of file to be copied
    :param string target_dir: Path that file is to be copied to
    :param string filename: Name of file to be copied
    :param bool overwrite: If true, overwrite if file already exists in target_dir
    :param bool delete_source: If true, delete file from source_dir after copying
    :return:
    """

    from google.cloud import storage
    storage_client = storage.Client()

    gcs_bucket_name, gcs_path = target_dir.split("/", 1)

    bucket = storage_client.get_bucket(gcs_bucket_name)

    blob = bucket.blob(gcs_path + "/" + filename)
    blob.upload_from_filename(source_dir + "/" + filename )

    if delete_source:
        os.remove(source_dir + "/" + filename)



def _make_local_path(path_name):
    if path_name[0] == ".":
        path_name = os.path.join(os.getcwd(), path_name)
        path_name = os.path.abspath(path_name)
    return path_name

def _s3_key_exists(s3, s3_bucket, s3_key):
    content = s3.head_object(Bucket=s3_bucket, Key=s3_key)
    if content.get('ResponseMetadata', None) is not None:
        return True
    else:
        return False


class DataBunch:
    """
    A class that groups together DataSets. A typical DataBunch would consist of "train", "valid" and "test" DataSets.
    """

    def __init__(self, data_sets_dict=None):
        if data_sets_dict:
            self.add_data_sets(data_sets_dict)

    def add_data_sets(self, data_sets_dict):
        """
        Adds datasets to this DataBunch. For example

        data_bunch = DataBunch()
        data_bunch.add_data_sets(
            {
                "train": train_data_set,
                "valid": valid_data_set,
                "test": test_data_set
            }
        )

        :param dict data_sets_dict: Of the form {{"data_set_name": data_set},...}
        :return:
        """

        for data_set_name, data_set in data_sets_dict.items():
            self.add_data_set(data_set_name, data_set)

    def add_data_set(self, data_set_name, data_set):
        """
        Adds a single DataSet to this DataBunch.

        :param string data_set_name: name of the DataSet.
        :param DataSet data_set: the DataSet to add.
        :return:
        """

        setattr(self, data_set_name, data_set)


    def transform(self, data_set_names, params, transform_then_slice=None):
        """
        Transforms all the underlying DataSets contained in the list data_set_names and return a new DataBunch with the
        transformed datasets

        :param list data_set_names: Names of the DataSets that should be transformed
        :param dict params: Parameters according to which the DataSets should be transformed
        :param bool transform_then_slice: Passed to DataSet.transform. Determines whether each DataWrapper within the DataSet should
        be transformed independently, or whether only "full_data" should be transformed and the subset the derived from there
        :return: new DataBunch
        """

        output_data_bunch = DataBunch()
        for data_set_name in data_set_names:
            input_data_set = getattr(self, data_set_name)
            output_data_bunch.add_data_set(
                data_set_name=data_set_name,
                data_set=input_data_set.transform(transform_then_slice=transform_then_slice,
                                                  transformation_params=params))

        return output_data_bunch




from keras import backend as K
import numpy as np
import pandas as pd
import os


def get_mock_loss_function(mock_value):
    """
    A loss function that just returns a hardcoded value. Used for testing only.

    :param mock_value:
    :return:
    """

    def mock_loss(y_true, y_pred):
        return mock_value
    return mock_loss

def get_weighted_loss_function(label_df_path, label_columns, label_df_read_params=None):
    """
    Returns a weighted_binary_crossentropy loss function. Calculates the weights based on an input DataFrame that can
    be read from a local disk location

    :param string label_df_path: The path where a DataFrame with the observed labels can be found
    :param list label_columns: The columns within the DataFrame where the observed labels are kept
    :param dict label_df_read_params: The parameters to be used to read the label DataFrame
    :return: callable weighted_binary_crossentropy
    """

    y_true = _load_label_df(label_df_path, label_columns, label_df_read_params)
    weights = _get_class_weights(y_true)
    def weighted_binary_crossentropy(y_true, y_pred):
        """
        A loss function that returns a weighted binary crossentropy. This will result in equal penalties being placed on
        all classes. Particularly useful when modeling mutliple binary labels (i.e. "attributes") where so labels are
        observed far less freuently than others

        :param y_true:
        :param y_pred:
        :return:
        """

        return K.mean((weights[:,0]**(1-y_true))*(weights[:,1]**(y_true))*K.binary_crossentropy(y_true, y_pred), axis=-1)
    return weighted_binary_crossentropy

def _get_class_weights(y_true):
    """
    Computes balanced class weighted based on the observed values
    """

    from sklearn.utils.class_weight import compute_class_weight
    number_dim = np.shape(y_true)[1]
    weights = np.empty([number_dim, 2])
    for i in range(number_dim):
        weights[i] = compute_class_weight("balanced", [0.,1.], y_true[:, i])
    return weights

def _load_label_df(label_df_path, label_columns, label_df_read_params=None):
    """
    Loads a label_df from a local path
    """
    extension = os.path.splitext(label_df_path)[1]
    if not label_df_read_params:
        label_df_read_params = {}
    if extension == ".pkl":
        label_df = pd.read_pickle(label_df_path, **label_df_read_params)
    elif extension == ".csv":
        label_df = pd.read_csv(label_df_path, **label_df_read_params)
    elif extension == ".json":
        label_df = pd.read_json(label_df_path, **label_df_read_params)
    else:
        raise NotImplementedError("Extension '{}' has not yet been implemented")

    return label_df[label_columns].values


"""
Simple IoC containers that provide direct access to various H2O providers
"""


class ModelDefinitions:

    from h2o.estimators.deeplearning import H2OAutoEncoderEstimator
    from h2o.estimators.deeplearning import H2ODeepLearningEstimator
    from h2o.estimators.deepwater import H2ODeepWaterEstimator
    from h2o.estimators.gbm import H2OGradientBoostingEstimator
    from h2o.estimators.glm import H2OGeneralizedLinearEstimator
    from h2o.estimators.glrm import H2OGeneralizedLowRankEstimator
    from h2o.estimators.kmeans import H2OKMeansEstimator
    from h2o.estimators.naive_bayes import H2ONaiveBayesEstimator
    from h2o.estimators.pca import H2OPrincipalComponentAnalysisEstimator
    from h2o.estimators.random_forest import H2ORandomForestEstimator
    from h2o.estimators.svd import H2OSingularValueDecompositionEstimator
    
    rf = H2ORandomForestEstimator
    mlp = H2ODeepLearningEstimator
    deep_water = H2ODeepWaterEstimator
    auto_encoder = H2OAutoEncoderEstimator
    gbm = H2OGradientBoostingEstimator
    glm = H2OGeneralizedLinearEstimator
    low_rank = H2OGeneralizedLowRankEstimator
    k_means = H2OKMeansEstimator
    naive_bayes = H2ONaiveBayesEstimator
    pca = H2OPrincipalComponentAnalysisEstimator
    svd = H2OSingularValueDecompositionEstimator


class ModelFitters:
    from mercury_ml.h2o.providers import model_fitting

    fit = model_fitting.fit

class ModelSavers:
    from mercury_ml.h2o.providers import model_saving

    save_h2o_model = model_saving.save_h2o_model
    save_json_details = model_saving.save_model_details
    save_pojo = model_saving.save_pojo
    save_mojo = model_saving.save_mojo
    save_pojo_jar = model_saving.save_pojo_jar
    save_mojo_jar = model_saving.save_mojo_jar

class ModelLoaders:
    from mercury_ml.h2o.providers import model_loading
    load_h2o_model = model_loading.load_h2o_model
    load_mojo = model_loading.load_mojo_model


class ModelEvaluators:
    from mercury_ml.h2o.providers import model_evaluation

    evaluate = model_evaluation.evaluate
    evaluate_threshold_metrics = model_evaluation.evaluate_threshold_metrics


class PredictionFunctions:
    from mercury_ml.h2o.providers import prediction

    predict = prediction.predict


class SessionInitiators:
    from mercury_ml.h2o.providers import session

    get_or_create_h2o = session.get_or_create_h2o
    get_or_create_h2o_sparkling = session.get_or_create_h2o_sparkling


def fit(model, data_bunch, callbacks, batch_size, epochs, class_weight=None, return_best_model=False,
        custom_objects=None):

    """
    Fits a Keras model when using arrays as inputs

    :param model: A (compiled) Keras model
    :param DataBunch data_bunch: A DataBunch consisting of "train" and "valid" DataSets
    :param dict callbacks: The Keras callbacks to perform during training
    :param int batch_size: The batch size to use during training
    :param int epochs: The number of Epochs to train for
    :param dict class_weight: Class weights to be applied during training
    :param bool return_best_model: If set to true, instead of return the model from the final epoch, the model with the
    best accuracy will be returned
    :param custom_objects: Any custom objects with which a best model may have been saved (custom loss functions would count
    as custom objects)
    :return: A (fitted) Keras model, the training history
    """

    train_x = data_bunch.train.features.underlying
    train_y = data_bunch.train.targets.underlying
    valid_x = data_bunch.valid.features.underlying
    valid_y = data_bunch.valid.targets.underlying


    try:
        model.fit(x=train_x,
                  y=train_y,
                  validation_data=(valid_x, valid_y),
                  batch_size=batch_size,
                  epochs=epochs,
                  class_weight=class_weight,
                  callbacks=callbacks
                  )

    except KeyboardInterrupt:
        print("Interrupted by user")
        pass

    if return_best_model:
        # by default Keras returns the model from the last epoch
        from keras.models import load_model
        model = load_model(__get_model_checkpoint(callbacks).filepath,
                           custom_objects=custom_objects)
        print("return_best_model set to True. Returning best model")
    else:
        print("return_best_model set to False. Returning model from last epoch")

    return model


def fit_generator(model, data_bunch, callbacks, epochs, class_weight=None, return_best_model=False,
        custom_objects=None):

    """
    Fits a Keras model when using generators as input

    :param model: A (compiled) Keras model
    :param DataBunch data_bunch: A DataBunch consisting of "train" and "valid" DataSets
    :param dict callbacks: The Keras callbacks to perform during training
    :param int batch_size: The batch size to use during training
    :param int epochs: The number of Epochs to train for
    :param dict class_weight: Class weights to be applied during training
    :param bool return_best_model: If set to true, instead of return the model from the final epoch, the model with the
    best accuracy will be returned
    :param custom_objects: Any custom objects with which a best model may have been saved (custom loss functions would count
    as custom objects)
    :return: A (fitted) Keras model, the training history
    """

    train_iterator = data_bunch.train.features.underlying
    valid_iterator = data_bunch.valid.features.underlying

    try:
        model.fit_generator(generator=train_iterator,
                                  validation_data=valid_iterator,
                                  steps_per_epoch= train_iterator.n // train_iterator.batch_size,
                                  validation_steps= valid_iterator.n // valid_iterator.batch_size,
                                  epochs=epochs,
                                  class_weight=class_weight,
                                  callbacks=callbacks
                                  )

    except KeyboardInterrupt:
        print("Interrupted by user")
        pass

    if return_best_model:
        # by default Keras returns the model from the last epoch
        from keras.models import load_model
        model = load_model(__get_model_checkpoint(callbacks).filepath,
                           custom_objects=custom_objects)
        print("return_best_model set to True. Returning best model")
    else:
        print("return_best_model set to False. Returning model from last epoch")

    return model


def __get_model_checkpoint(callbacks):
    return [callback for callback in callbacks if callback.__class__.__name__=="ModelCheckpointProvider"][0]


import os

def copy_from_gcs_to_disk(source_dir, target_dir, filename, overwrite=False, delete_source=False):
    """
    Moves a file from GCS to Disk

    :param string source_dir: Path of file to be copied
    :param string target_dir: Path that file is to be copied to
    :param string filename: Name of file to be copied
    :param bool overwrite: If true, overwrite if file already exists in target_dir
    :param bool delete_source: If true, delete file from source_dir after copying
    :return:
    """

    if not overwrite and os.path.isfile(target_dir + "/" + filename):
        pass
    else:
        from google.cloud import storage
        storage_client = storage.Client()

        gcs_bucket_name, gcs_path = target_dir.split("/", 1)

        bucket = storage_client.get_bucket(gcs_bucket_name)

        blob = bucket.blob(gcs_path + "/" + filename)
        blob.download_to_filename(source_dir + "/" + filename)

        if delete_source:
            blob.delete()




import json
import os

from h2o.utils.typechecks import assert_is_type
from h2o.model.model_base import ModelBase, H2OValueError
import h2o


def save_h2o_model(model, filename, extension=None, local_dir=None, force=False):
    """
    Adapts the "h2o.save_model" method in the h2o module in order to control the name of the saved model

    :param model: An H2O Model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param string extension: Extension to the filename with which the model is to be saved
    :param bool force: If true will overwrite existing files
    :return: The full filepath where the model is saved
    """


    if not extension:
        extension = ""

    assert_is_type(model, ModelBase)
    assert_is_type(local_dir, str)
    assert_is_type(force, bool)

    local_dir = os.getcwd() if not local_dir else _make_local_path(local_dir)
    print(local_dir)

    
    filename = filename + extension
    path = os.path.join(local_dir, filename)

    _make_dirs(local_dir)

    final_path = h2o.api("GET /99/Models.bin/%s" % model.model_id, data={"dir": path, "force": force})["dir"]

    return final_path


def save_pojo(model, filename, extension=None, local_dir=None):
    """
    Adapts the "h2o.download_pojo" method in the h2o module in order to control the name of the saved model, and to
    split "pojo" and "pojo_jar" downloads into two different functions

    :param model: An H2O Model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param string extension: Extension to the filename with which the model is to be saved
    :return: The full filepath where the model is saved
    """

    if not extension:
        extension = ".java"

    assert_is_type(model, ModelBase)

    local_dir = os.getcwd() if not local_dir else _make_local_path(local_dir)
    filename = filename + extension
    path = os.path.join(local_dir, filename)

    _make_dirs(local_dir)

    if not model.have_pojo:
        raise H2OValueError("Export to POJO not supported")

    final_path = h2o.api("GET /3/Models.java/%s" % model.model_id, save_to=path)

    return final_path


def save_pojo_jar(model, filename, extension=None, local_dir=None):
    """
    adapts the "h2o.download_pojo" method in the h2o module in order to control the name of the saved model, and to
    split "pojo" and "pojo_jar" downloads into two different functions

    :param model: An H2O Model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param string extension: Extension to the filename with which the model is to be saved
    :return: The full filepath where the model is saved
    """

    if not extension:
        extension = ".jar"

    assert_is_type(model, ModelBase)

    if not model.have_pojo:
        raise H2OValueError("Export to POJO not supported")

    local_dir = os.getcwd() if not local_dir else _make_local_path(local_dir)
    filename = filename + extension
    path = os.path.join(local_dir, filename)

    _make_dirs(local_dir)

    final_path = h2o.api("GET /3/h2o-genmodel.jar", save_to=path)

    return final_path


def save_mojo(model, filename, extension=None, local_dir=None, force=False):
    """
    adapts the "model_base.ModelBase.save_mojo" method in the h2o module in order to control the name of the saved model, and to
    split "mojo" and "mojo_jar" downloads into two different functions

    :param model: An H2O Model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param string extension: Extension to the filename with which the model is to be saved
    :param bool force: If true will overwrite existing files
    :return: The full filepath where the model is saved
    """


    if not extension:
        extension = ".zip"

    assert_is_type(local_dir, str)
    assert_is_type(force, bool)

    if not model.have_mojo:
        raise H2OValueError("Export to MOJO not supported")

    local_dir = os.getcwd() if not local_dir else _make_local_path(local_dir)
    filename = filename + extension
    path = os.path.join(local_dir, filename)

    _make_dirs(local_dir)

    final_path = h2o.api("GET /99/Models.mojo/%s" % model.model_id, data={"dir": path, "force": force})["dir"]

    return final_path


def save_mojo_jar(model, filename, extension=None, local_dir=None, force=False):
    """
    adapts the "model_base.ModelBase.save_mojo" method in the h2o module in order to control the name of the saved model, and to
    split "mojo" and "mojo_jar" downloads into two different functions

    :param model: An H2O Model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param string extension: Extension to the filename with which the model is to be saved
    :param bool force: If true will overwrite existing files
    :return: The full filepath where the model is saved
    """



    if not extension:
        extension = ".jar"

    assert_is_type(local_dir, str)
    assert_is_type(force, bool)

    if not model.have_mojo:
        raise H2OValueError("Export to MOJO not supported")

    local_dir = os.getcwd() if not local_dir else _make_local_path(local_dir)
    filename = filename + extension
    path = os.path.join(local_dir, filename)

    _make_dirs(local_dir)

    final_path = h2o.api("GET /3/h2o-genmodel.jar", save_to=path)

    return final_path

def save_model_details(model, filename, extension=None, local_dir=None, force=False):
    """
    adapts the "model_base.ModelBase.save_model_details" method in the h2o module in order to control the name of the saved model

    :param model: An H2O Model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param string extension: Extension to the filename with which the model is to be saved
    :param bool force: If true will overwrite existing files
    :return: The full filepath where the model is saved
    """

    if not extension:
        extension = ".jar"


    assert_is_type(local_dir, str)
    assert_is_type(force, bool)

    local_dir = os.getcwd() if not local_dir else _make_local_path(local_dir)
    filename = filename + extension
    path = os.path.join(local_dir, filename)

    _make_dirs(local_dir)

    final_path = h2o.api("GET /99/Models/%s/json" % model.model_id, data={"dir": path, "force": force})["dir"]

    return final_path

def _make_dirs(dir):
    import os
    if not os.path.exists(dir):
        os.makedirs(dir)

def _make_local_path(path_name):
    if path_name[0] == ".":
        path_name = os.path.join(os.getcwd(), path_name)
        path_name = os.path.abspath(path_name)
    return path_name


class NumpyDataWrapper():
    """
    A DataWrapper with a Numpy array as its underlying data
    """

    def __init__(self, underlying, field_names):
        self.underlying = underlying
        self.field_names = field_names

    def slice_on_column_names(self, column_names):
        """
        Returns a Numpy array consisting of the columns that are present in the column_names list

        :param list list column_names: the names of the columns to return
        :return: Numpy array consisting of the selected columns
        """

        column_number_list =self._get_field_name_indices(column_names)
        return self.underlying[:, column_number_list]

    def to_h2o(self, factor_columns_list=None):
        """
        Creates a new H2ODataWrapper based on this NumpyDataWrapper and returns it as a new object

        :return: a new H2ODataWrapper
        """
        return self.to_pandas().to_h2o(factor_columns_list)

    def to_pandas(self):
        """
        Creates a new PandasDataWrapper based on this NumpyDataWrapper and returns it as a new object

        :return: a new PandasDataWrapper
        """
        import pandas as pd
        from mercury_ml.common.providers.data_wrappers.pandas import PandasDataWrapper
        df = pd.DataFrame(data=self.underlying, columns=self.field_names)

        return PandasDataWrapper(df, self.field_names)

    def to_spark(self):
        """
        Creates a new SparkDataWrapper based on this NumpyDataWrapper and returns it as a new object

        :return: a new SparkDataWrapper
        """
        return self.to_pandas().to_spark()

    def to_h2o_sparkling(self, spark_session_params=None, h2o_sparkling_params=None):
        """
        Creates a new H2OSparklingDataWrapper based on this NumpyDataWrapper and returns it as a new object

        :return: a new H2OSparklingDataWrapper
        """
        return self.to_pandas().to_h2o_sparkling(spark_session_params, h2o_sparkling_params)

    def to_numpy(self):
        return self

    def _get_field_name_indices(self, field_names_subset):
        """
        Takes a list of field names, compares if with self.field_names and return a list of the indices where field_names_subset
        intersects field self.field_names
        :param list field_names_subset: A list of strings with field names
        :return: A list of integers (indices)
        """
        ls = []
        for field_name in field_names_subset:
            ls = ls + [self.field_names.index(field_name)]
        return ls



from sklearn.metrics import roc_auc_score
from sklearn.metrics import confusion_matrix
from sklearn.metrics import accuracy_score
from copy import copy
import numpy as np


def evaluate_auc(y_true, y_pred, labels):
    """
    Calculates the AUC for each each of the individual observed targets / predictions within y_true / y_pred and returns
    a formatted dictionary with the results.

    :param ndarray y_true: The (potentially multiple) observed targets
    :param ndarray y_pred: The (potentially multiple) prediction probabilities
    :param list labels: Labels for which metric should be calculated
    :return: A dictionary of the form {label_metric_name: {label_name, label_metric_value}}
    """

    i = 0
    auc_dict = {}
    for label in labels:
        try:
            auc_dict[label] = np.asscalar(roc_auc_score(y_true=y_true[:, i].astype("float"),
                                                        y_score=y_pred[:, i].astype("float")))
        except:
            auc_dict[label] = None

        i = i + 1

    dict_to_return = {}
    dict_to_return["AUC"] = auc_dict
    return dict_to_return


def evaluate_confusion_matrix(y_true, y_pred, labels):
    """
    Calculates a Confusion for each each of for y_true / y_pred and returns a formatted dictionary with the results.

    :param ndarray y_true: The (potentially multiple) observed targets
    :param ndarray y_pred: The (potentially multiple) prediction probabilities
    :param list labels: Labels for which metric should be calculated
    :return: A dictionary of the form {label_metric_name: {label_name, label_metric_value}}
    """

    y_true_argmax = y_true.argmax(axis=-1)
    y_pred_argmax = y_pred.argmax(axis=-1)

    matrix = confusion_matrix(y_pred=y_pred_argmax,
                              y_true=y_true_argmax)

    cm_count_dict = {}
    cm_rate_dict = {}
    dict_to_return = {}
    for col in range(len(labels)):
        for row in range(len(labels)):
            cm_count_dict[labels[row]] = np.asscalar(matrix[row, col])
            cm_rate_dict[labels[row]] = np.asscalar(matrix[row, col] / matrix[row, :].sum())
        dict_to_return["ConfMat_Count_" + labels[col]] = copy(cm_count_dict)
        dict_to_return["ConfMat_Rate_" + labels[col]] = copy(cm_rate_dict)

    return dict_to_return



def evaluate_accuracy(y_true, y_pred, labels):
    """
    Calculates the Accuracy for each each of the individual observed targets / predictions within y_true / y_pred and returns
    a formatted dictionary with the results.

    :param ndarray y_true: The (potentially multiple) observed targets
    :param ndarray y_pred: The (potentially multiple) prediction probabilities
    :param list labels: Labels for which metric should be calculated
    :return: A dictionary of the form {label_metric_name: {label_name, label_metric_value}}
    """

    i = 0
    acc_dict = {}

    y_pred = y_pred.astype("float")
    y_true = y_true.astype("float")

    for label in labels:
        acc_dict[label] = np.asscalar(accuracy_score(y_true=_apply_threshold(y_true[:, i]),
                                                     y_pred=_apply_threshold(y_pred[:, i])))
        i = i + 1

    dict_to_return = {}
    dict_to_return["Accuracy"] = acc_dict
    return dict_to_return



def evaluate_multi_label_confusion_matrix(y_true, y_pred, labels):
    """
    Calculates binary confusion matrices for each each of the individual observed targets / predictions within y_true / y_pred and returns
    a formatted dictionary with the results.

    :param ndarray y_true: The (potentially multiple) observed targets
    :param ndarray y_pred: The (potentially multiple) prediction probabilities
    :param list labels: Labels for which metric should be calculated
    :return: A dictionary of the form {label_metric_name: {label_name, label_metric_value}}
    """

    dict_to_return = {"ConfMat_Count__00": {},
                      "ConfMat_Count__01": {},
                      "ConfMat_Count__10": {},
                      "ConfMat_Count__11": {},
                      "ConfMat_Rate__00": {},
                      "ConfMat_Rate__01": {},
                      "ConfMat_Rate__10": {},
                      "ConfMat_Rate__11": {}
                      }

    y_pred = y_pred.astype("float")
    y_true = y_true.astype("float")

    for label_col in range(len(labels)):
        y_true_flat = _apply_threshold(y_true[:, label_col])
        y_pred_flat = _apply_threshold(y_pred[:, label_col])
        matrix = confusion_matrix(y_pred=y_pred_flat, y_true=y_true_flat)
        for col in [0, 1]:
            for row in [0, 1]:
                count_dict = {labels[label_col]: np.asscalar(matrix[row, col])}
                rate_dict = {labels[label_col]: np.asscalar(matrix[row, col] / matrix[row, :].sum())}
                dict_to_return["ConfMat_Count__" + str(row) + str(col)].update(count_dict)
                dict_to_return["ConfMat_Rate__" + str(row) + str(col)].update(rate_dict)

    return dict_to_return




def _apply_threshold(array, threshold=0.5):
    thresholded_array = np.copy(array)
    thresholded_array[thresholded_array >= threshold] = 1
    thresholded_array[thresholded_array < threshold] = 0
    return thresholded_array



from sklearn.metrics import roc_auc_score


def evaluate_macro_auc(y_true, y_pred):
    """
    Calculates a the marco AUC for the (potentially multi-target) observations vs the predicted probabilities

    :param array y_true: The observed targets
    :param array y_pred: The prediction probabilities
    :return: A dictionary of the form {metric_name:metric_value}
    """

    return roc_auc_score(y_true=y_true.astype("float"), y_score=y_pred.astype("float"), average="macro")

def evaluate_micro_auc(y_true, y_pred):
    """
    Calculates a the micro AUC for the (potentially multi-target) observations vs the predicted probabilities

    :param array y_true: The observed targets
    :param array y_pred: The prediction probabilities
    :return: A dictionary of the form {metric_name:metric_value}
    """

    return roc_auc_score(y_true=y_true.astype("float"), y_score=y_pred.astype("float"), average="micro")



#TODO possibly split into two modules: data_set readers and data_bunch readers

import pandas as pd

from mercury_ml.common.providers.data_set import DataSet
from mercury_ml.common.providers.data_wrappers.pandas import PandasDataWrapper
import os

def read_pandas_data_set(path, input_format, full_data_columns, index_columns, features_columns, targets_columns):
    """
    Reads a pandas dataset from a local source and creates a DataSet consisting of PandasDataWrappers for full_data
    index, features and targets

    :param string path: The local path from which the Pandas DataFrame should be read.
    :param string input_format: The format (e.g. .csv) of the input data.
    :param list full_data_columns: The full list of columns that should be read from the input file.
    :param list index_columns: A subset of full_data_columns. The columns corresponding to the unique index.
    :param list features_columns: A subset of full_data_columns. The columns corresponding to the features.
    :param list targets_columns: A subset of full_data_columns. The columns corresponding to the targets.
    :return: DataSet consiting of PandasDataWrappers
    """

    if input_format == ".csv":
        df = pd.read_csv(path, usecols=full_data_columns)
    else:
        raise NotImplementedError("input_format '{}' not implemented yet")


    if input_format == ".pkl":
        df = pd.read_pickle(path)[full_data_columns]
    elif input_format == ".csv":
        df = pd.read_csv(path, usecols=full_data_columns)
    elif input_format == ".json":
        df = pd.read_json(path)[full_data_columns]
    else:
        raise NotImplementedError("Extension '{}' has not yet been implemented")

    return DataSet(
                {
                    "full_data": PandasDataWrapper(df, full_data_columns),
                    "index": PandasDataWrapper(df[index_columns], index_columns),
                    "features": PandasDataWrapper(df[features_columns], features_columns),
                    "targets": PandasDataWrapper(df[targets_columns], targets_columns)
                }
    )


def read_keras_single_input_image_iterator_data_set(generator_params, iterator_params):
    """
    Reads a Keras SingleInputDirectoryInterator from a local source and creates a DataSet consisting of full_dta, targets and index

    :param dict generator_params: parameters to initialise the SingleInputImageDataGenerator
    :param dict iterator_params: parameters to pass to the SingleInputImageDataGenerator.flow_from_directory in order to get the SingleInputDirectoryInterator

    :return: DataSet consiting of KerasIteratorFeaturesDataWrapper, KerasIteratorTargetsDataWrapper and KerasIteratorIndexDataWrapper
    """

    from mercury_ml.keras.providers.image_generators.single_input import SingleInputImageDataGenerator
    from mercury_ml.common.providers.data_wrappers.keras import KerasIteratorFeaturesDataWrapper, \
        KerasIteratorTargetsDataWrapper, KerasIteratorIndexDataWrapper

    generator = SingleInputImageDataGenerator(**generator_params)
    iterator = generator.flow_from_directory(**iterator_params)

    return DataSet(
                {
                    "features": KerasIteratorFeaturesDataWrapper(iterator, None),
                    "targets": KerasIteratorTargetsDataWrapper(iterator, None),
                    "index": KerasIteratorIndexDataWrapper(iterator, None)
                }
            )


def read_keras_multi_label_image_iterator_data_set(
        label_df_path, index_column, label_columns, generator_params, iterator_params, label_df_read_params=None):
    """
    Reads a Keras MultiLabelDirectoryInterator from a local source and creates a DataSet consisting of full_dta, targets and index

    :param string label_df_path: A local path where a DataFrame that has the multi-label information can be found
    :param list index_column: The name of the column that make up the unique index
    :param list label_columns: The names of the columns that contain the label information
    :param dict generator_params: parameters to initialise the MultiLabelImageDataGenerator
    :param dict iterator_params: parameters to pass to the MultiLabelImageDataGenerator.flow_from_directory in order to get the MultiLabelDirectoryInterator
    :param dict DataFrame label_df_read_params: The parameters according to which the label DataFrame should be read
    :return:
    """

    from mercury_ml.keras.providers.image_generators.multi_label import MultiLabelImageDataGenerator
    from mercury_ml.common.providers.data_wrappers.keras import KerasIteratorFeaturesDataWrapper, \
        KerasIteratorTargetsDataWrapper, KerasIteratorIndexDataWrapper

    extension = os.path.splitext(label_df_path)[1]
    if not label_df_read_params:
        label_df_read_params = {}
    if extension == ".pkl":
        label_df = pd.read_pickle(label_df_path, **label_df_read_params)
    elif extension == ".csv":
        label_df = pd.read_csv(label_df_path, **label_df_read_params)
    elif extension == ".json":
        label_df = pd.read_json(label_df_path, **label_df_read_params)
    else:
        raise NotImplementedError("Extension '{}' has not yet been implemented")

    columns_to_select = [index_column] + label_columns
    # label_df must consist entirely of label columns. Therefore we set an explicit index. This will be used to
    # identify the image that the labels belong to
    label_df = label_df[columns_to_select].set_index(index_column)

    generator = MultiLabelImageDataGenerator(label_df, **generator_params)
    iterator = generator.flow_from_directory(**iterator_params)

    return DataSet(
                {
                    "features": KerasIteratorFeaturesDataWrapper(iterator, None),
                    "targets": KerasIteratorTargetsDataWrapper(iterator, None),
                    "index": KerasIteratorIndexDataWrapper(iterator, None)
                }
            )


def load_referenced_json_config(filepath):
    """
    Loads a json file that has "jsonref" style references and resolves the references
    :param string filepath: The path from which to reads the JSON
    :return: dict
    """
    import jsonref
    from copy import deepcopy
    with open(filepath, "r") as f:
        config = jsonref.load(f)

    # this dereferences the dictionary produced by jsonref.load. See here https://github.com/gazpachoking/jsonref/issues/9
    config = deepcopy(config)
    return config

def recursively_update_config(config, string_formatting_dict):
    """
    Recursively updates a dictionary using string formatting.

    For example:
    > config = {
    >    "filename": "{model_purpose}_{session_id}_hdf5.h5",
    >    "save_path": "./results/{model_purpose}"
    > }
    >
    > string_formatting_dict = {"model_purpose": "test", "model_id": "12345"}
    >
    > recursively_update_config(config, string_formatting_dict)
    > print(config)
    {"filename": "test_12345_hdf5.h5", "save_path": "./results/test"}

    :param iterable config: A dictionary or list to be updated
    :param dict string_formatting_dict: A dictionary containing the information with which "config" is to be formatted
    :return:
    """

    for k in _iterate_list_or_dict(config):
        v = config[k]
        if isinstance(v, dict) or isinstance(v, list):
            recursively_update_config(v, string_formatting_dict)
        else:
            if _key_in_string(v, string_formatting_dict):
                config[k] = v.format(**string_formatting_dict)

def _iterate_list_or_dict(obj):
    """Iterates over object if list or dictionary"""
    return obj if isinstance(obj, dict) else range(len(obj))


def _key_in_string(string, string_formatting_dict):
    """Checks which formatting keys are present in a given string"""
    key_in_string = False
    if isinstance(string, str):
        for key, value in string_formatting_dict.items():
            if "{" + key + "}" in string:
                key_in_string = True
    return key_in_string


from setuptools import setup, find_packages
setup(name="mercury-ml",
      version="0.1.1",
      description="A library for managing Machine Learning workflows",
      url="https://github.com/mercury-ml-team/mercury-ml",
      author="Karl Schriek",
      author_email="kschriek@gmail.com",
      license="MIT",
      packages=find_packages(
          exclude=["*.tests", "*.tests.*", "tests.*", "tests",
                   "*.examples", "*.examples.*", "examples.*", "examples"]),
      include_package_data=True,
      install_requires=["numpy", "pandas", "scikit-learn", "jsonref"],
      extras_require={
            "keras": ["tensorflow", "keras", "pillow"],
            "keras-gpu": ["tensorflow-gpu", "keras", "pillow"],
            "h2o": ["h2o"],
            "h2o-sparkling": ["h2o", "pyspark", "h2o-pysparkling"],
            "s3": ["boto3"],
            "gcs": ["google-cloud-storage"]
            },
      python_requires=">=3.5",
      zip_safe=False)






from keras.callbacks import Callback, TensorBoard, BaseLogger, EarlyStopping, ModelCheckpoint, TerminateOnNaN, \
        ProgbarLogger, RemoteMonitor, LearningRateScheduler, ReduceLROnPlateau, CSVLogger

import os

# from ml_workflow.utils import utils
# from ml_workflow.containers.core import Core
# from sklearn.metrics import roc_auc_score
#
# class AUCGeneratorCallBackProvider(Callback):
#     """ Class for AUC callback in Keras """
#     def __init__(self, params):
#         super(AUCGeneratorCallBackProvider, self).__init__()
#         self.generator = Core.session.datasets()[params["dataset_to_use"]]["data_formats"]["ml"][params["generator_to_use"]]
#         self.steps = params["steps"]
#         self.verbose = params["verbose"]
#         self.dataset = params["dataset_to_use"]
#
#     def on_epoch_end(self, epoch, logs={}):
#         """
#         A function that defines a Callback that should be run at the end of a training epoch
#         :param epoch:
#         :param logs:
#         :return:
#         """
#         # Get probability matrix
#         predictions = self.model.predict_generator(
#             generator=self.generator,
#             steps=self.steps,
#             verbose=self.verbose)
#
#         # Target matrix
#
#         targets = self.generator.get_target_dummies()[:self.generator.batch_size*self.steps,:]
#         # macro average AUC
#         try:
#             auc_macro = roc_auc_score(y_true=targets, y_score=predictions, average="macro")
#             print(" - {}_auc: ".format(self.dataset) + str(round(auc_macro, 4)))
#         except:
#             print(" - {}_auc: calculation failed")
#

class ModelCheckpointProvider(ModelCheckpoint):
    """ Class for AUC callback in Keras """
    def __init__(self, params):
        super(ModelCheckpointProvider, self).__init__(**params)
        model_dir = os.path.dirname(params["filepath"])
        model_dir = self._make_local_path(model_dir)
        if not os.path.isdir(model_dir):
            os.makedirs(model_dir)

    def _make_local_path(self, path_name):
        if path_name[0] == ".":
            path_name = os.path.join(os.getcwd(), path_name)
            path_name = os.path.abspath(path_name)
        return path_name

class TensorBoardProvider(TensorBoard):
    """ Class for AUC callback in Keras """
    def __init__(self, params):
        super(TensorBoardProvider, self).__init__(**params)

class BaseLoggerProvider(BaseLogger):
    """ Class for AUC callback in Keras """
    def __init__(self, params):
        super(BaseLoggerProvider, self).__init__(**params)


class EarlyStoppingProvider(EarlyStopping):
    """ Class for AUC callback in Keras """
    def __init__(self, params):
        super(EarlyStoppingProvider, self).__init__(**params)


class TerminateOnNaNProvider(TerminateOnNaN):
    """ Class for AUC callback in Keras """
    def __init__(self, params):
        super(TerminateOnNaNProvider, self).__init__()


class ProgbarLoggerProvider(ProgbarLogger):
    """ Class for AUC callback in Keras """
    def __init__(self, params):
        super(ProgbarLoggerProvider, self).__init__(**params)


class CSVLoggerProvider(CSVLogger):
    """ Class for AUC callback in Keras """
    def __init__(self, params):
        super(CSVLoggerProvider, self).__init__(**params)


class ReduceLROnPlateauProvider(ReduceLROnPlateau):
    """ Class for AUC callback in Keras """
    def __init__(self, params):
        super(ReduceLROnPlateauProvider, self).__init__(**params)


class LearningRateSchedulerProvider(LearningRateScheduler):
    """ Class for AUC callback in Keras """
    def __init__(self, params):
        super(LearningRateSchedulerProvider, self).__init__(**params)


class RemoteMonitorProvider(RemoteMonitor):
    """ Class for AUC callback in Keras """
    def __init__(self, params):
        super(RemoteMonitorProvider, self).__init__(**params)



def evaluate(model, data_set):
    """
    Fetches the built-in Keras metrics, when using an array as input
    :param model: A trainied Keras model
    :param data_set: A DataSet with NumpyDataWrappers called "features" and "targets"
    :return: A dictionary of the form {metric_name:metric_value,...}
    """

    metric_values = model.evaluate(x=data_set.features.underlying, y=data_set.targets.underlying)
    metrics = __generate_metric_dict(model.metrics_names, metric_values)

    return metrics

def evaluate_generator(model, data_set):
    """
    Fetches the built-in Keras metrics, when using an generator as input
    :param model: A trainied Keras model
    :param data_set: A DataSet with a KerasIteratorFeaturesDataWrapper called "features" and a KerasIteratorTargetsDataWrapper called "targets"
    :return: A dictionary of the form {metric_name:metric_value,...}
    """

    iterator = data_set.features.underlying
    metric_values = model.evaluate_generator(iterator)
    metrics = __generate_metric_dict(model.metrics_names, metric_values)

    return metrics


def __generate_metric_dict(metric_names, metric_values):

    metrics = {}
    for i in range(len(metric_names)):
        metrics[metric_names[i]] = metric_values[i]

    return metrics


from keras.initializers import glorot_normal
from keras.layers import Conv2D, MaxPooling2D, Dense, Flatten, Dropout
from keras.models import Sequential

def define_model(input_size, dropout_rate, final_activation, nb_classes, seed=1234):
    """
    A small example function for defining a simple (convolutional) model. This just serves to illustrate what a "define_model"
    function might look like. It is expected that users will create their own "define_model" functions to use together with
    the other mercury-ml functions

    :param list input_size: the shape of the expected input data.
    :param double dropout_rate: The dropout rate.
    :param string final_activation: The activation function to use in the final layer.
    :param int nb_classes: The number of classes.
    :param int seed: The random seed.
    :return: A (defined) Keras model.
    """

    dense_initializer = glorot_normal(seed=seed)
    conv_initializer = glorot_normal(seed=seed)

    model = Sequential()
    model.add(Conv2D(4, kernel_size=(5, 5), strides=(2, 2),
                     kernel_initializer=conv_initializer,
                     input_shape=(input_size[0],
                                  input_size[1], 3)))

    model.add(MaxPooling2D(pool_size=(3, 3), strides=(2, 2)))
    model.add(Flatten())
    model.add(Dense(2, activation="relu",
                    kernel_initializer=dense_initializer))
    model.add(Dropout(rate=dropout_rate, seed=seed))

    model.add(Dense(nb_classes, activation=final_activation))

    return model



#TODO possibly split into two modules: data_set readers and data_bunch readers

from mercury_ml.common.providers.data_bunch import DataBunch
from mercury_ml.common.providers.data_set import DataSet
from mercury_ml.common.providers.data_wrappers.spark import SparkDataWrapper

def read_spark_data_set(source_table, where_clause, full_data_columns, index_columns, features_columns, targets_columns):
    from mercury_ml.h2o.providers.session import get_or_create_spark
    spark = get_or_create_spark() #TODO this should be only "get" at this stage
    spark_df = spark.sql(
        "SELECT * FROM {} WHERE {}".format(source_table, where_clause)
    )

    return DataSet(
                {
                    "full_data": SparkDataWrapper(spark_df.select(*full_data_columns), full_data_columns),
                    "index": SparkDataWrapper(spark_df.select(*index_columns), index_columns),
                    "features": SparkDataWrapper(spark_df.select(*features_columns), features_columns),
                    "targets": SparkDataWrapper(spark_df.select(*targets_columns), targets_columns)
                }
    )



def predict(data_set, model, **kwargs):
    """
    Produces predictions with a trained Keras model where inputs are arrays

    :param DataSet data_set: A DataSet with a NumpyDataWrapper called "features"
    :param model: A (fitted) Keras model
    :param kwargs: Additional parameters to be passed to model.predict
    :return: A NumpyDataWrapper with an array of predictions as its underlying
    """

    prediction_array = model.predict(x=data_set.features.underlying, **kwargs)
    from mercury_ml.common.providers.data_wrappers.numpy import NumpyDataWrapper
    return NumpyDataWrapper(underlying=prediction_array, field_names=data_set.targets.field_names)


def predict_generator(data_set, model, **kwargs):
    """
    Produces predictions with a trained Keras model where inputs are generators

    :param DataSet data_set: A DataSet with a KerasIteratorFeaturesDataWrapper called "features"
    :param model: A (fitted) Keras model
    :param kwargs: Additional parameters to be passed to model.predict
    :return: A NumpyDataWrapper with an array of predictions as its underlying
    """

    prediction_array = model.predict_generator(generator=data_set.features.underlying, **kwargs)
    from mercury_ml.common.providers.data_wrappers.numpy import NumpyDataWrapper
    return NumpyDataWrapper(underlying=prediction_array, field_names=data_set.features.underlying.get_labels_dummies())